library regression;
Uses Math;
{$mode objfpc}{$H+}

type
  {MT4 stores the data of each candle in such a structure.
  MT4 will then give us a pointer to an array of such candles}
  TCandle = packed record
    time:   LongInt; // POSIX timestamp
    open,
    high,
    low,
    close,
    volume: Double;
  end;

  {the following would be a pretty large array but we actually never create
  an instance of it, we only need a typed pointer to such an array.
  We will later receive a pointer from MT4 to the chart history
  stored in such an array}
  THistory = array[0..MaxLongInt div SizeOf(TCandle) - 1] of TCandle;
  PHistory = ^THistory;

  {This is the structure of an indicator buffer. We will be given a
  pointer to an indicator buffer to fill it with the calculated values}
  TBuffer = array[0..MaxLongInt div SizeOf(Double) - 1] of Double;
  PBuffer = ^TBuffer;

  TVal = array[0..2] of Double;
  PVal = ^TVal;

  var i0: LongInt = 0;


const
  EMPTY_VALUE = $7FFFFFFF;


{calculate the moving average for the bar at index. Index is the absolute index
in the history array of the bar for which we want to calculate this value. The
function will return the value or it will return EMPTY_VALUE if there are not
enough bars before index to calculate a moving average of the requested period}
function movingAverage(history: PHistory; period: LongInt; degree: LongInt; index: LongInt; var vals: PVal): Double;
var
  sum: Double;
  i:   LongInt;
  ai:  array[0..9,0..9] of double;
  b:   array[0..9] of double;
  x:   array[0..9] of double;
  sx:  array[0..19] of double;
  ip:  LongInt;
   p:  LongInt;
   n:  LongInt;
   f:  LongInt;
  ii:  LongInt;
  jj:  LongInt;
  kk:  LongInt;
  ll:  LongInt;
  nn:  LongInt;
  sq:  Double;
  qq:  Double;
  mm:  Double;
  tt:  Double;
  mi:  LongInt;
  fx:  array of Double;


begin
  if index < period then exit(EMPTY_VALUE);
  if period < 1 then exit(history^[index].Close);


  sum := 0;
  ip:=period;
  p:=ip;
  sx[1] :=p+1;
  nn := degree+1;
  SetLength(fx,p+1);

  mi := 1;
  for mi:=1  to nn*2-2 do
  begin
  sum:=0;
    for n:=i0 to i0+p do
    begin
         sum := sum+ power(n,mi);
    end;
  sx[mi+1]:=sum;
  end;


  for mi:=1 to nn  do
  begin
       sum:=0.00000;
       for n:=i0 to i0+p do
       begin
         if (mi = 1) then sum:= sum+ history^[index-n].Close
         else sum:= sum+ (history^[index-n].Close * power(n,mi-1));
       end;
       b[mi]:=sum;
  end;

  for jj:=1 to nn do
  begin
    for ii:=1 to nn do
    begin
       kk:=ii+jj-1;
       ai[ii,jj]:=sx[kk];
    end
  end;

  for kk:=1 to nn-1 do
  begin
       ll:=0;
       mm:=0;
       for ii:=kk to nn do
       begin
            if (Abs(ai[ii,kk])>mm) then
            begin
                mm:=Abs(ai[ii,kk]);
                ll:=ii;
            end;
       end;

       if ll=0 then exit(EMPTY_VALUE) else
       begin
            if ll<>kk then
            begin
                 for jj:=1 to nn do
                 begin
                      tt:=ai[kk,jj];
                      ai[kk,jj]:=ai[ll,jj];
                      ai[ll,jj]:=tt;
                 end;
            tt:=b[kk];
            b[kk]:=b[ll];
            b[ll]:=tt;
            end;
       end;
       for ii:=kk+1 to nn do
       begin
            qq:=ai[ii,kk]/ai[kk,kk];
            for jj:=1 to nn do
            begin
            if jj=kk then ai[ii,jj]:=0
            else ai[ii,jj]:=ai[ii,jj]-qq*ai[kk,jj];
            end;
       b[ii]:=b[ii]-qq*b[kk];
       end;
  end;
  x[nn]:=b[nn]/ai[nn,nn];
  for ii:=nn-1 Downto 1 do
  begin
      tt:=0;
      for jj:=1 to nn-ii do
      begin
           tt:=tt+ai[ii,ii+jj]*x[ii+jj];
           x[ii]:=(1/ai[ii,ii])*(b[ii]-tt);
      end;

  end;


  for n:=i0 to i0+p do
  begin
    sum:=0;
    for kk:=1 to degree do
    begin
       sum:=sum+x[kk+1]*power(n,kk);
    end;
    fx[n]:=x[1]+sum;

  end;

  sq:=0.0;
  for n:=i0 to i0+p do
  begin
       sq:=sq+power(history^[index-n].Close-fx[n],2);
  end;
  sq:=sqrt(sq/(p+1));
  vals^[0] := fx[0];
  vals^[1] := fx[1];
  vals^[2] :=sq;
end;

{Calculate for all (new) bars the indicator values and update the
indicator buffer. This is more or less the same thing that is found
in the example indicators from MetaQuotes. On every tick we look at
the numbers from Bars and IndicatorCounted() to determine over which
bars we have to loop and (re-)calculate the indicator value, only
here we do it in Pascal and it is nicely documented and the code
looks much cleaner}
procedure updateBuffer(history: PHistory; bars: LongInt; degree: LongInt; period: LongInt; vals: PVal); stdcall;
var
  i:  LongInt;
  iC: LongInt;
  iB: LongInt;

begin
  // unlike in metatrader which silently tranforms the
  // indices of the buffers so that in mql4 the 0 is always
  // the last bar in the buffer, the real underlying arrays
  // with which we are dealing here directly start with 0
  // as the oldest bar and bars - 1 as the current bar.

  // We tranform bars and indicator_counted from MT4 into
  // abolute indices in our buffer arrays so that
  // iB will be the index of the current bar and
  // iC will be the index of the last seen bar
  // in most cases they will be identical and
  // we loop only one iteration: the current bar.
  // When a new bar has just appeared we will loop
  // over the last two bars.

  movingAverage(history, period, degree, bars-1, vals);


end;

exports
  updateBuffer;

begin
end.
